@@ -4,6 +4,7 @@
 #include <string.h>
 
 #include <orbit/orbit.h>
+#include <orbit/orb-core/orb-types.h>
 
 #include "../orbit-init.h"
 #include "../poa/orbit-poa-export.h"
@@ -40,14 +41,20 @@ static gint         orbit_initial_recv_limit = -1;
 static gboolean     orbit_use_irda           = FALSE;
 static gboolean     orbit_use_ssl            = FALSE;
 static gboolean     orbit_use_genuid_simple  = FALSE;
+#ifdef G_OS_WIN32
+static gboolean     orbit_local_only         = TRUE;
+#else
 static gboolean     orbit_local_only         = FALSE;
+#endif
+static char        *orbit_net_id             = NULL;
 static gboolean     orbit_use_http_iors      = FALSE;
 static char        *orbit_ipsock             = NULL;
-static char        *orbit_ipname             = NULL;
+static const char  *orbit_ipname             = NULL;
 static char        *orbit_debug_options      = NULL;
 static char        *orbit_naming_ref         = NULL;
 static GSList      *orbit_initref_list       = NULL; 
 static gboolean     orbit_use_corbaloc       = FALSE;
+static guint        orbit_timeout_msec       = 60000; /* 60 seconds - 0 will disable timeouts altogether */
 
 void
 ORBit_ORB_start_servers (CORBA_ORB orb)
@@ -66,7 +73,36 @@ ORBit_ORB_start_servers (CORBA_ORB orb)
 		create_options |= LINK_CONNECTION_LOCAL_ONLY;
 
 	if (orbit_local_only || (orbit_use_usocks && !(orbit_use_ipv4 || orbit_use_ipv6 || orbit_use_irda || orbit_use_ssl)))
-			link_use_local_hostname (TRUE);
+			link_use_local_hostname (LINK_NET_ID_IS_LOCAL);
+	else {
+		do {
+			if (!orbit_net_id)
+				break;
+
+			if (!strcmp(orbit_net_id, "local")) {
+				link_use_local_hostname (LINK_NET_ID_IS_LOCAL);
+				break;
+			}
+			if (!strcmp(orbit_net_id, "short")) {
+				link_use_local_hostname (LINK_NET_ID_IS_SHORT_HOSTNAME);
+				break;
+			}
+			if (!strcmp(orbit_net_id, "fqdn")) {
+				link_use_local_hostname (LINK_NET_ID_IS_FQDN);
+				break;
+			}
+			if (!strcmp(orbit_net_id, "ipaddr")) {
+				link_use_local_hostname (LINK_NET_ID_IS_IPADDR);
+				break;
+			}
+			link_set_local_hostname(orbit_net_id);
+		} while (0);
+	}
+		    
+	if (!orbit_ipname)
+		orbit_ipname = link_get_local_hostname();
+	else
+		link_set_local_hostname(orbit_ipname);
 
 	for (info = link_protocol_all (); info->name; info++) {
 		GIOPServer           *server;
@@ -113,8 +149,11 @@ static void
 strip_object_profiles (gpointer o, gpointer b, gpointer c)
 {
 	CORBA_Object obj = o;
+
 	IOP_delete_profiles (obj->orb, &obj->profile_list);
 	IOP_delete_profiles (obj->orb, &obj->forward_locations);
+
+	obj->orb = NULL;
 }
 
 static void
@@ -245,6 +284,7 @@ static CORBA_ORB _ORBit_orb = NULL;
 static gulong    init_level = 0;
 static gboolean  atexit_shutdown = FALSE;
 
+#ifndef G_OS_WIN32 /* See comment at g_atexit() call below */
 /*
  *   This is neccessary to clean up any remaining UDS
  * and to flush any remaining oneway traffic in buffers.
@@ -270,6 +310,7 @@ shutdown_orb (void)
 
 	atexit_shutdown = FALSE;
 }
+#endif
 
 static
 gboolean
@@ -376,8 +417,8 @@ CORBA_ORB_init (int *argc, char **argv,
 	}
 #endif /* G_ENABLE_DEBUG */
 
-
 	giop_recv_set_limit (orbit_initial_recv_limit);
+	giop_set_timeout (orbit_timeout_msec);
 	giop_init (thread_safe,
 		   orbit_use_ipv4 || orbit_use_ipv6 ||
 		   orbit_use_irda || orbit_use_ssl);
@@ -398,11 +439,39 @@ CORBA_ORB_init (int *argc, char **argv,
 	/* released by CORBA_ORB_destroy */
 	_ORBit_orb = ORBit_RootObject_duplicate (retval);
 	_ORBit_orb->lock = link_mutex_new ();
+#ifndef G_OS_WIN32
+	/* atexit(), which g_atexit() is just a #define for on Win32,
+	 * often causes breakage when invoked from DLLs. It causes the
+	 * registered function to be called when the calling DLL is
+	 * being unloaded. At that time, however, random other DLLs
+	 * might also have already been unloaded. There is no
+	 * guarantee WinSock even works any longer. Etc. Best to avoid
+	 * atexit() completely on Win32, and hope that just exiting
+	 * the process and thus severing all connections will be
+	 * noticed by all peers the process was connected to and acted
+	 * upon properly.
+	 *
+	 * In the evolution-exchange-storage process's case, the
+	 * shutdown_orb() function caused the process to hang and not
+	 * exit, leaving the sockets it was listening on still in a
+	 * LISTEN state. bonobo-activation-server thought the bonobo
+	 * servers in evolution-exchange-storage were still OK and
+	 * tried to contact them when Evolution was started the next
+	 * time, causing it to hang, too.
+	 */
 	g_atexit (shutdown_orb);
+#endif
 
 	retval->default_giop_version = GIOP_LATEST;
 
 	retval->adaptors = g_ptr_array_new ();
+
+	/* init the forward bind hashtable*/
+	retval->forw_binds = g_hash_table_new_full (
+		g_str_hash, g_str_equal,
+		g_free,
+		NULL); 
+
 	ORBit_init_internals (retval, ev);
 	/* FIXME, handle exceptions */ 
 
@@ -492,7 +561,9 @@ CORBA_ORB_string_to_object (CORBA_ORB          orb,
 	CORBA_Object         retval = CORBA_OBJECT_NIL;
 	CORBA_unsigned_long  len;
 	GIOPRecvBuffer      *buf;
+#if defined ENABLE_HTTP
 	gchar               *ior = NULL;
+#endif
 	guchar              *tmpbuf;
 	int                  i;
 
@@ -508,9 +579,19 @@ CORBA_ORB_string_to_object (CORBA_ORB          orb,
 		    strstr (string, "://")) {
 			/* FIXME: this code is a security hazard */
 			ior = orb_http_resolve (string);
-			if (!ior)
-				return CORBA_OBJECT_NIL;
+			if (!ior) {
+				/* FIXME, set error minor code
+				 * (vendor's error code) to tell user
+				 * initial location of error, ie my
+				 * local ORB, proxy's ORB, server's
+				 * ORB, etc. */
+				CORBA_exception_set_system (
+					ev,
+					ex_CORBA_BAD_PARAM,
+					CORBA_COMPLETED_NO);
 
+				return CORBA_OBJECT_NIL;
+			}
 			string = ior;
 		} else 
 #endif
@@ -532,9 +613,9 @@ CORBA_ORB_string_to_object (CORBA_ORB          orb,
 			len--;
 		
 		if (len % 2) {
-			if (ior)
-				g_free (ior);
-			
+#if defined ENABLE_HTTP
+			g_free (ior);
+#endif
 			return CORBA_OBJECT_NIL;
 		}
 		
@@ -556,10 +637,9 @@ CORBA_ORB_string_to_object (CORBA_ORB          orb,
 		}
 		
 		giop_recv_buffer_unuse (buf);
-		
-		if (ior)
-			g_free (ior);
-		
+#if defined ENABLE_HTTP		
+		g_free (ior);
+#endif
 		return retval;
 	} else {
 		return ORBit_object_by_corbaloc (orb, string, ev);
@@ -572,13 +652,13 @@ CORBA_ORB_create_list (CORBA_ORB          obj,
 		       CORBA_NVList      *new_list,
 		       CORBA_Environment *ev)
 {
-	CORBA_NVList new;
+	CORBA_NVList nvlist;
 
-	new = g_new0 (struct CORBA_NVList_type, 1);
+	nvlist = g_new0 (struct CORBA_NVList_type, 1);
 
-	new->list = g_array_new (FALSE, TRUE, sizeof (CORBA_NamedValue));
+	nvlist->list = g_array_new (FALSE, TRUE, sizeof (CORBA_NamedValue));
 
-	*new_list = new;
+	*new_list = nvlist;
 }
 
 void
@@ -1241,8 +1321,17 @@ CORBA_ORB_destroy (CORBA_ORB          orb,
 		}
 	}
 
+	/* destroy the forward bind hashtable*/
+	g_hash_table_destroy (orb->forw_binds);
+	orb->forw_binds = NULL;
+	
 	orb->life_flags |= ORBit_LifeF_Destroyed;
 
+	if (orb->lock) {
+		g_mutex_free (orb->lock);
+		orb->lock = NULL;
+	}
+
 	ORBit_RootObject_release (orb);
 
 	/* At this stage there should be 1 ref left in the system -
@@ -1311,7 +1400,13 @@ ORBit_ORB_forw_bind (CORBA_ORB                   orb,
 		     CORBA_Object                obj,
 		     CORBA_Environment          *ev)
 {
-	g_warning ("ORBit_ORB_forw_bind NYI");
+	
+	if (obj)
+		g_hash_table_insert (orb->forw_binds, objkey->_buffer, obj);
+	else {
+		g_hash_table_remove(orb->forw_binds, objkey->_buffer);
+	}
+	
 }
 
 gboolean
@@ -1328,6 +1423,25 @@ ORBit_proto_use (const char *name)
 	return FALSE;
 }
 
+/**
+ * ORBit_get_giop_recv_limit:
+ * 
+ * This function will return the GIOP receive limit. The
+ * GIOP receive limit is the maximum number of bytes that
+ * are allowed be received in any one ingoing GIOP 
+ * communication. This function is essential if an application
+ * is about to receive a big amount of data. Knowing the GIOP 
+ * receive limit will enable the application to poll to data
+ * in chunks that are below the receive limit.
+ *
+ * Since: 2.14.1
+ */
+glong
+ORBit_get_giop_recv_limit (void)
+{
+	return giop_recv_get_limit ();
+}
+
 const ORBit_option orbit_supported_options[] = {
 	{ "ORBid",              ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
 	{ "ORBImplRepoIOR",     ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
@@ -1335,10 +1449,11 @@ const ORBit_option orbit_supported_options[] = {
 	{ "ORBNamingIOR",       ORBIT_OPTION_STRING,  &orbit_naming_ref},
 
 	{ "ORBRootPOAIOR",      ORBIT_OPTION_STRING,  NULL }, /* FIXME: huh?          */
- 	{ "ORBIIOPIPName",      ORBIT_OPTION_STRING,  &orbit_ipname },
+ 	{ "ORBIIOPIPName",      ORBIT_OPTION_STRING,  &orbit_ipname }, /* Will always take precedence over ORBNetID */
  	{ "ORBIIOPIPSock",      ORBIT_OPTION_STRING,  &orbit_ipsock },
 	{ "ORBInitialMsgLimit", ORBIT_OPTION_INT,     &orbit_initial_recv_limit },
 	{ "ORBLocalOnly",       ORBIT_OPTION_BOOLEAN, &orbit_local_only },
+ 	{ "ORBNetID",           ORBIT_OPTION_STRING,  &orbit_net_id },
 	/* warning: this option is a security risk unless used with LocalOnly */
 	{ "ORBIIOPIPv4",        ORBIT_OPTION_BOOLEAN, &orbit_use_ipv4 },
 	{ "ORBIIOPIPv6",        ORBIT_OPTION_BOOLEAN, &orbit_use_ipv6 },
@@ -1353,5 +1468,81 @@ const ORBit_option orbit_supported_options[] = {
 	{ "ORBDebugFlags",      ORBIT_OPTION_STRING,  &orbit_debug_options },
 	{ "ORBInitRef",         ORBIT_OPTION_KEY_VALUE,  &orbit_initref_list},
 	{ "ORBCorbaloc",        ORBIT_OPTION_BOOLEAN, &orbit_use_corbaloc},
-	{ NULL,                 0,                    NULL },
+	{ "GIOPTimeoutMSEC",    ORBIT_OPTION_ULONG,   &orbit_timeout_msec },
+	{ NULL,                 0,                    NULL }
 };
+
+#ifdef G_OS_WIN32
+
+/* DllMain function used to store the DLL handle */
+static HMODULE hmodule;
+G_LOCK_DEFINE_STATIC (mutex);
+
+/* Silence gcc warnings about no prototype */
+BOOL WINAPI DllMain (HINSTANCE hinstDLL,
+		     DWORD     fdwReason,
+		     LPVOID    lpvReserved);
+const gchar *ORBit_win32_get_typelib_dir (void);
+const gchar *ORBit_win32_get_system_rcfile (void);
+
+BOOL WINAPI
+DllMain (HINSTANCE hinstDLL,
+	 DWORD     fdwReason,
+	 LPVOID    lpvReserved)
+{
+        switch (fdwReason) {
+        case DLL_PROCESS_ATTACH:
+                hmodule = hinstDLL;
+                break;
+        }
+        return TRUE;
+}
+
+static const char *typelib_dir = NULL;
+static const char *system_rcfile = NULL;
+
+static void
+setup (void)
+{
+	gchar *prefix;
+
+	G_LOCK (mutex);
+	if (typelib_dir != NULL) {
+		G_UNLOCK (mutex);
+		return;
+	}
+
+	prefix = g_win32_get_package_installation_directory_of_module (hmodule);
+
+	if (prefix == NULL) {
+		/* Just to not crash... */
+		prefix = g_strdup ("");
+	}
+
+	typelib_dir = g_strconcat (prefix,
+				   "\\lib\\orbit-2.0",
+				   NULL);
+
+	system_rcfile = g_strconcat (prefix,
+				     "\\etc\\orbitrc",
+				     NULL);
+	G_UNLOCK (mutex);
+
+	g_free (prefix);
+}
+
+const gchar *
+ORBit_win32_get_typelib_dir (void)
+{
+	setup ();
+	return typelib_dir;
+}
+
+const gchar *
+ORBit_win32_get_system_rcfile (void)
+{
+	setup ();
+	return system_rcfile;
+}
+
+#endif
